@@ -5,8 +5,6 @@
  *	Copyright (C) 2002 ActiveState Corporation
  *	Copyright (C) 2004 Starfish Systems 
  *
- * $Header: /cvsroot/tls/tls/tls.c,v 1.35 2014/12/08 19:09:06 andreas_kupries Exp $
- *
  * TLS (aka SSL) Channel - can be layered on any bi-directional
  * Tcl_Channel (Note: Requires Trf Core Patch)
  *
@@ -40,33 +38,34 @@
 		Tcl_TranslateFileName(interp, (key), (dsp)))
 #define REASON()	ERR_reason_error_string(ERR_get_error())
 
-static void	InfoCallback _ANSI_ARGS_ ((CONST SSL *ssl, int where, int ret));
+static void	InfoCallback(CONST SSL *ssl, int where, int ret);
 
-static int	CiphersObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	CiphersObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static int	HandshakeObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	HandshakeObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static int	ImportObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	ImportObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static int	StatusObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	StatusObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static int	VersionObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	VersionObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static int	MiscObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	MiscObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static int	UnimportObjCmd _ANSI_ARGS_ ((ClientData clientData,
-			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+static int	UnimportObjCmd(ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
 
-static SSL_CTX *CTX_Init _ANSI_ARGS_((State *statePtr, int proto, char *key,
-			char *cert, char *CAdir, char *CAfile, char *ciphers));
+static SSL_CTX *CTX_Init(State *statePtr, int proto, char *key,
+			char *cert, char *CAdir, char *CAfile, char *ciphers,
+			char *DHparams);
 
-static int	TlsLibInit _ANSI_ARGS_ (()) ;
+static int	TlsLibInit(int uninitialize);
 
 #define TLS_PROTO_SSL2		0x01
 #define TLS_PROTO_SSL3		0x02
@@ -79,42 +78,10 @@ static int	TlsLibInit _ANSI_ARGS_ (()) ;
  * Static data structures
  */
 
-#ifndef NO_DH
-/* from openssl/apps/s_server.c */
-
-static unsigned char dh512_p[]={
-	0xDA,0x58,0x3C,0x16,0xD9,0x85,0x22,0x89,0xD0,0xE4,0xAF,0x75,
-	0x6F,0x4C,0xCA,0x92,0xDD,0x4B,0xE5,0x33,0xB8,0x04,0xFB,0x0F,
-	0xED,0x94,0xEF,0x9C,0x8A,0x44,0x03,0xED,0x57,0x46,0x50,0xD3,
-	0x69,0x99,0xDB,0x29,0xD7,0x76,0x27,0x6B,0xA2,0xD3,0xD4,0x12,
-	0xE2,0x18,0xF4,0xDD,0x1E,0x08,0x4C,0xF6,0xD8,0x00,0x3E,0x7C,
-	0x47,0x74,0xE8,0x33,
-	};
-static unsigned char dh512_g[]={
-	0x02,
-};
-
-static DH *get_dh512()
-{
-    DH *dh=NULL;
-
-    if ((dh=DH_new()) == NULL) return(NULL);
-
-    dh->p=BN_bin2bn(dh512_p,sizeof(dh512_p),NULL);
-    dh->g=BN_bin2bn(dh512_g,sizeof(dh512_g),NULL);
-
-    if ((dh->p == NULL) || (dh->g == NULL))
-	return(NULL);
-    return(dh);
-}
+#ifndef OPENSSL_NO_DH
+#include "dh_params.h"
 #endif
 
-/*
- * Defined in Tls_Init to determine what kind of channels we are using
- * (old-style 8.2.0-8.3.1 or new-style 8.3.2+).
- */
-int channelTypeVersion;
-
 /*
  * We lose the tcl password callback when we use the RSA BSAFE SSL-C 1.1.2
  * libraries instead of the current OpenSSL libraries.
@@ -150,27 +117,38 @@ int channelTypeVersion;
  * Based from /crypto/cryptlib.c of OpenSSL and NSOpenSSL.
  */
 
-static Tcl_Mutex locks[CRYPTO_NUM_LOCKS];
+static Tcl_Mutex *locks = NULL;
+static int locksCount = 0;
 static Tcl_Mutex init_mx;
-static int initialized;
 
-static void          CryptoThreadLockCallback (int mode, int n, const char *file, int line);
-static unsigned long CryptoThreadIdCallback   (void);
+void CryptoThreadLockCallback(int mode, int n, const char *file, int line) {
 
-static void
-CryptoThreadLockCallback(int mode, int n, const char *file, int line)
-{
-    if (mode & CRYPTO_LOCK) {
-       Tcl_MutexLock(&locks[n]);
-    } else {
-       Tcl_MutexUnlock(&locks[n]);
-    }
+	if (mode & CRYPTO_LOCK) {
+		/* This debugging is turned off by default -- it's too noisy. */
+		/* dprintf("Called to lock (n=%i of %i)", n, locksCount); */
+		Tcl_MutexLock(&locks[n]);
+	} else {
+		/* dprintf("Called to unlock (n=%i of %i)", n, locksCount); */
+		Tcl_MutexUnlock(&locks[n]);
+	}
+
+	/* dprintf("Returning"); */
+
+	return;
+	file = file;
+	line = line;
 }
 
-static unsigned long
-CryptoThreadIdCallback(void)
-{
-    return (unsigned long) Tcl_GetCurrentThread();
+unsigned long CryptoThreadIdCallback(void) {
+	unsigned long ret;
+
+	dprintf("Called");
+
+	ret = (unsigned long) Tcl_GetCurrentThread();
+
+	dprintf("Returning %lu", ret);
+
+	return(ret);
 }
 #endif /* OPENSSL_THREADS */
 #endif /* TCL_THREADS */
@@ -197,6 +175,8 @@ InfoCallback(CONST SSL *ssl, int where, int ret)
     Tcl_Obj *cmdPtr;
     char *major; char *minor;
 
+    dprintf("Called");
+
     if (statePtr->callback == (Tcl_Obj*)NULL)
 	return;
 
@@ -292,13 +272,13 @@ VerifyCallback(int ok, X509_STORE_CTX *ctx)
     Tcl_Obj *cmdPtr, *result;
     char *errStr, *string;
     int length;
-    SSL   *ssl		= (SSL*)X509_STORE_CTX_get_app_data(ctx);
+    SSL   *ssl		= (SSL*)X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
     X509  *cert		= X509_STORE_CTX_get_current_cert(ctx);
     State *statePtr	= (State*)SSL_get_app_data(ssl);
     int depth		= X509_STORE_CTX_get_error_depth(ctx);
     int err		= X509_STORE_CTX_get_error(ctx);
 
-    dprintf(stderr, "Verify: %d\n", ok);
+    dprintf("Verify: %d", ok);
 
     if (!ok) {
 	errStr = (char*)X509_verify_cert_error_string(err);
@@ -347,7 +327,7 @@ VerifyCallback(int ok, X509_STORE_CTX *ctx)
 	result = Tcl_GetObjResult(statePtr->interp);
 	string = Tcl_GetStringFromObj(result, &length);
 	/* An empty result leaves verification unchanged.	*/
-	if (length > 0) {
+	if (string != NULL && length > 0) {
 	    if (Tcl_GetIntFromObj(statePtr->interp, result, &ok) != TCL_OK) {
 		Tcl_BackgroundError(statePtr->interp);
 		ok = 0;
@@ -382,6 +362,8 @@ Tls_Error(State *statePtr, char *msg)
 {
     Tcl_Obj *cmdPtr;
 
+    dprintf("Called");
+
     if (msg && *msg) {
 	Tcl_SetErrorCode(statePtr->interp, "SSL", msg, (char *)NULL);
     } else {
@@ -440,6 +422,9 @@ static int
 PasswordCallback(char *buf, int size, int verify)
 {
     return -1;
+    	buf = buf;
+	size = size;
+	verify = verify;
 }
 #else
 static int
@@ -450,6 +435,8 @@ PasswordCallback(char *buf, int size, int verify, void *udata)
     Tcl_Obj *cmdPtr;
     int result;
 
+    dprintf("Called");
+
     if (statePtr->password == NULL) {
 	if (Tcl_EvalEx(interp, "tls::password", -1, TCL_EVAL_GLOBAL)
 		== TCL_OK) {
@@ -483,6 +470,7 @@ PasswordCallback(char *buf, int size, int verify, void *udata)
     } else {
 	return -1;
     }
+    	verify = verify;
 }
 #endif
 
@@ -522,6 +510,8 @@ CiphersObjCmd(clientData, interp, objc, objv)
     char *cp, buf[BUFSIZ];
     int index, verbose = 0;
 
+    dprintf("Called");
+
     if (objc < 2 || objc > 3) {
 	Tcl_WrongNumArgs(interp, 1, objv, "protocol ?verbose?");
 	return TCL_ERROR;
@@ -616,6 +606,7 @@ CiphersObjCmd(clientData, interp, objc, objv)
 
     Tcl_SetObjResult( interp, objPtr);
     return TCL_OK;
+    	clientData = clientData;
 }
 
 /*
@@ -635,64 +626,72 @@ CiphersObjCmd(clientData, interp, objc, objv)
  *-------------------------------------------------------------------
  */
 
-static int
-HandshakeObjCmd(clientData, interp, objc, objv)
-    ClientData clientData;	/* Not used. */
-    Tcl_Interp *interp;
-    int objc;
-    Tcl_Obj *CONST objv[];
-{
-    Tcl_Channel chan;		/* The channel to set a mode on. */
-    State *statePtr;		/* client state for ssl socket */
-    int ret = 1;
+static int HandshakeObjCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
+	Tcl_Channel chan;		/* The channel to set a mode on. */
+	State *statePtr;		/* client state for ssl socket */
+	CONST char *errStr = NULL;
+	int ret = 1;
+	int err = 0;
 
-    if (objc != 2) {
-	Tcl_WrongNumArgs(interp, 1, objv, "channel");
-	return TCL_ERROR;
-    }
+	dprintf("Called");
+
+	if (objc != 2) {
+		Tcl_WrongNumArgs(interp, 1, objv, "channel");
+		return(TCL_ERROR);
+	}
+
+	chan = Tcl_GetChannel(interp, Tcl_GetStringFromObj(objv[1], NULL), NULL);
+	if (chan == (Tcl_Channel) NULL) {
+		return(TCL_ERROR);
+	}
 
-    chan = Tcl_GetChannel(interp, Tcl_GetStringFromObj(objv[1], NULL), NULL);
-    if (chan == (Tcl_Channel) NULL) {
-	return TCL_ERROR;
-    }
-    if (channelTypeVersion == TLS_CHANNEL_VERSION_2) {
 	/*
 	 * Make sure to operate on the topmost channel
 	 */
 	chan = Tcl_GetTopChannel(chan);
-    }
-    if (Tcl_GetChannelType(chan) != Tls_ChannelType()) {
-	Tcl_AppendResult(interp, "bad channel \"", Tcl_GetChannelName(chan),
-		"\": not a TLS channel", NULL);
-	return TCL_ERROR;
-    }
-    statePtr = (State *)Tcl_GetChannelInstanceData(chan);
-
-    if (!SSL_is_init_finished(statePtr->ssl)) {
-	int err;
-	ret = Tls_WaitForConnect(statePtr, &err);
-	if ((statePtr->flags & TLS_TCL_ASYNC) && err == EAGAIN) {
-	    ret = 0;
+	if (Tcl_GetChannelType(chan) != Tls_ChannelType()) {
+		Tcl_AppendResult(interp, "bad channel \"", Tcl_GetChannelName(chan), "\": not a TLS channel", NULL);
+		return(TCL_ERROR);
 	}
-	if (ret < 0) {
-	    CONST char *errStr = statePtr->err;
-	    Tcl_ResetResult(interp);
-	    Tcl_SetErrno(err);
+	statePtr = (State *)Tcl_GetChannelInstanceData(chan);
 
-	    if (!errStr || *errStr == 0) {
-		errStr = Tcl_PosixError(interp);
-	    }
+	dprintf("Calling Tls_WaitForConnect");
+	ret = Tls_WaitForConnect(statePtr, &err, 1);
+	dprintf("Tls_WaitForConnect returned: %i", ret);
 
-	    Tcl_AppendResult(interp, "handshake failed: ", errStr,
-		    (char *) NULL);
-	    return TCL_ERROR;
+	if (
+	    ret < 0 && \
+	    ((statePtr->flags & TLS_TCL_ASYNC) && err == EAGAIN)
+	) {
+		dprintf("Async set and err = EAGAIN");
+		ret = 0;
+	} else if (ret < 0) {
+		errStr = statePtr->err;
+		Tcl_ResetResult(interp);
+		Tcl_SetErrno(err);
+
+		if (!errStr || *errStr == 0) {
+			errStr = Tcl_PosixError(interp);
+		}
+
+		Tcl_AppendResult(interp, "handshake failed: ", errStr, (char *) NULL);
+		dprintf("Returning TCL_ERROR with handshake failed: %s", errStr);
+		return(TCL_ERROR);
+	} else {
+		if (err != 0) {
+			dprintf("Got an error with a completed handshake: err = %i", err);
+		}
+
+		ret = 1;
 	}
-    }
 
-    Tcl_SetObjResult(interp, Tcl_NewIntObj(ret));
-    return TCL_OK;
+	dprintf("Returning TCL_OK with data \"%i\"", ret);
+	Tcl_SetObjResult(interp, Tcl_NewIntObj(ret));
+	return(TCL_OK);
+
+    	clientData = clientData;
 }
-
+
 /*
  *-------------------------------------------------------------------
  *
@@ -723,6 +722,7 @@ ImportObjCmd(clientData, interp, objc, objv)
     SSL_CTX *ctx	= NULL;
     Tcl_Obj *script	= NULL;
     Tcl_Obj *password	= NULL;
+    Tcl_DString upperChannelTranslation, upperChannelBlocking, upperChannelEncoding, upperChannelEOFChar;
     int idx, len;
     int flags		= TLS_TCL_INIT;
     int server		= 0;	/* is connection incoming or outgoing? */
@@ -731,26 +731,34 @@ ImportObjCmd(clientData, interp, objc, objv)
     char *ciphers	= NULL;
     char *CAfile	= NULL;
     char *CAdir		= NULL;
+    char *DHparams	= NULL;
     char *model		= NULL;
 #ifndef OPENSSL_NO_TLSEXT
     char *servername	= NULL;	/* hostname for Server Name Indication */
 #endif
-#if defined(NO_SSL2)
-    int ssl2 = 0;
-#else
-    int ssl2 = 1;
-#endif
-#if defined(NO_SSL3)
-    int ssl3 = 0;
-#else
-    int ssl3 = 1;
-#endif
-    int tls1 = 1;
-    int tls1_1 = 1;
-    int tls1_2 = 1;
+    int ssl2 = 0, ssl3 = 0;
+    int tls1 = 1, tls1_1 = 1, tls1_2 = 1;
     int proto = 0;
     int verify = 0, require = 0, request = 1;
 
+    dprintf("Called");
+
+#if defined(NO_TLS1) && defined(NO_TLS1_1) && defined(NO_TLS1_2) && defined(NO_SSL3) && !defined(NO_SSL2)
+    ssl2 = 1;
+#endif
+#if defined(NO_TLS1) && defined(NO_TLS1_1) && defined(NO_TLS1_2) && defined(NO_SSL2) && !defined(NO_SSL3)
+    ssl3 = 1;
+#endif
+#if defined(NO_TLS1)
+    tls1 = 0;
+#endif
+#if defined(NO_TLS1_1)
+    tls1_1 = 0;
+#endif
+#if defined(NO_TLS1_2)
+    tls1_2 = 0;
+#endif
+
     if (objc < 2) {
 	Tcl_WrongNumArgs(interp, 1, objv, "channel ?options?");
 	return TCL_ERROR;
@@ -760,12 +768,11 @@ ImportObjCmd(clientData, interp, objc, objv)
     if (chan == (Tcl_Channel) NULL) {
 	return TCL_ERROR;
     }
-    if (channelTypeVersion == TLS_CHANNEL_VERSION_2) {
-	/*
-	 * Make sure to operate on the topmost channel
-	 */
-	chan = Tcl_GetTopChannel(chan);
-    }
+
+    /*
+     * Make sure to operate on the topmost channel
+     */
+    chan = Tcl_GetTopChannel(chan);
 
     for (idx = 2; idx < objc; idx++) {
 	char *opt = Tcl_GetStringFromObj(objv[idx], NULL);
@@ -778,6 +785,7 @@ ImportObjCmd(clientData, interp, objc, objv)
 	OPTSTR( "-certfile", cert);
 	OPTSTR( "-cipher", ciphers);
 	OPTOBJ( "-command", script);
+	OPTSTR( "-dhparams", DHparams);
 	OPTSTR( "-keyfile", key);
 	OPTSTR( "-model", model);
 	OPTOBJ( "-password", password);
@@ -794,7 +802,7 @@ ImportObjCmd(clientData, interp, objc, objv)
 	OPTBOOL( "-tls1.1", tls1_1);
 	OPTBOOL( "-tls1.2", tls1_2);
 
-	OPTBAD( "option", "-cadir, -cafile, -certfile, -cipher, -command, -keyfile, -model, -password, -require, -request, -server, -servername, -ssl2, -ssl3, -tls1, -tls1.1 or -tls1.2");
+	OPTBAD( "option", "-cadir, -cafile, -certfile, -cipher, -command, -dhparams, -keyfile, -model, -password, -require, -request, -server, -servername, -ssl2, -ssl3, -tls1, -tls1.1 or -tls1.2");
 
 	return TCL_ERROR;
     }
@@ -809,11 +817,12 @@ ImportObjCmd(clientData, interp, objc, objv)
     proto |= (tls1_2 ? TLS_PROTO_TLS1_2 : 0);
 
     /* reset to NULL if blank string provided */
-    if (cert && !*cert)		cert	= NULL;
-    if (key && !*key)		key	= NULL;
-    if (ciphers && !*ciphers)	ciphers	= NULL;
-    if (CAfile && !*CAfile)	CAfile	= NULL;
-    if (CAdir && !*CAdir)	CAdir	= NULL;
+    if (cert && !*cert)		cert	 = NULL;
+    if (key && !*key)		key	 = NULL;
+    if (ciphers && !*ciphers)	ciphers	 = NULL;
+    if (CAfile && !*CAfile)	CAfile	 = NULL;
+    if (CAdir && !*CAdir)	CAdir	 = NULL;
+    if (DHparams && !*DHparams)	DHparams = NULL;
 
     /* new SSL state */
     statePtr		= (State *) ckalloc((unsigned) sizeof(State));
@@ -850,12 +859,11 @@ ImportObjCmd(clientData, interp, objc, objv)
 	    Tls_Free((char *) statePtr);
 	    return TCL_ERROR;
 	}
-	if (channelTypeVersion == TLS_CHANNEL_VERSION_2) {
-	    /*
-	     * Make sure to operate on the topmost channel
-	     */
-	    chan = Tcl_GetTopChannel(chan);
-	}
+
+        /*
+         * Make sure to operate on the topmost channel
+         */
+        chan = Tcl_GetTopChannel(chan);
 	if (Tcl_GetChannelType(chan) != Tls_ChannelType()) {
 	    Tcl_AppendResult(interp, "bad channel \"",
 		    Tcl_GetChannelName(chan), "\": not a TLS channel", NULL);
@@ -864,8 +872,8 @@ ImportObjCmd(clientData, interp, objc, objv)
 	}
 	ctx = ((State *)Tcl_GetChannelInstanceData(chan))->ctx;
     } else {
-	if ((ctx = CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers))
-		== (SSL_CTX*)0) {
+	if ((ctx = CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers,
+		DHparams)) == (SSL_CTX*)0) {
 	    Tls_Free((char *) statePtr);
 	    return TCL_ERROR;
 	}
@@ -879,19 +887,19 @@ ImportObjCmd(clientData, interp, objc, objv)
      * We only want to adjust the buffering in pre-v2 channels, where
      * each channel in the stack maintained its own buffers.
      */
+    Tcl_DStringInit(&upperChannelTranslation);
+    Tcl_DStringInit(&upperChannelBlocking);
+    Tcl_DStringInit(&upperChannelEOFChar);
+    Tcl_DStringInit(&upperChannelEncoding);
+    Tcl_GetChannelOption(interp, chan, "-eofchar", &upperChannelEOFChar);
+    Tcl_GetChannelOption(interp, chan, "-encoding", &upperChannelEncoding);
+    Tcl_GetChannelOption(interp, chan, "-translation", &upperChannelTranslation);
+    Tcl_GetChannelOption(interp, chan, "-blocking", &upperChannelBlocking);
     Tcl_SetChannelOption(interp, chan, "-translation", "binary");
-    if (channelTypeVersion == TLS_CHANNEL_VERSION_1) {
-	Tcl_SetChannelOption(interp, chan, "-buffering", "none");
-    }
-
-    if (channelTypeVersion == TLS_CHANNEL_VERSION_2) {
-	statePtr->self = Tcl_StackChannel(interp, Tls_ChannelType(),
-		(ClientData) statePtr, (TCL_READABLE | TCL_WRITABLE), chan);
-    } else {
-	statePtr->self = chan;
-	Tcl_StackChannel(interp, Tls_ChannelType(),
-		(ClientData) statePtr, (TCL_READABLE | TCL_WRITABLE), chan);
-    }
+    Tcl_SetChannelOption(interp, chan, "-blocking", "true");
+    dprintf("Consuming Tcl channel %s", Tcl_GetChannelName(chan));
+    statePtr->self = Tcl_StackChannel(interp, Tls_ChannelType(), (ClientData) statePtr, (TCL_READABLE | TCL_WRITABLE), chan);
+    dprintf("Created channel named %s", Tcl_GetChannelName(statePtr->self));
     if (statePtr->self == (Tcl_Channel) NULL) {
 	/*
 	 * No use of Tcl_EventuallyFree because no possible Tcl_Preserve.
@@ -900,6 +908,11 @@ ImportObjCmd(clientData, interp, objc, objv)
 	return TCL_ERROR;
     }
 
+    Tcl_SetChannelOption(interp, statePtr->self, "-translation", Tcl_DStringValue(&upperChannelTranslation));
+    Tcl_SetChannelOption(interp, statePtr->self, "-encoding", Tcl_DStringValue(&upperChannelEncoding));
+    Tcl_SetChannelOption(interp, statePtr->self, "-eofchar", Tcl_DStringValue(&upperChannelEOFChar));
+    Tcl_SetChannelOption(interp, statePtr->self, "-blocking", Tcl_DStringValue(&upperChannelBlocking));
+
     /*
      * SSL Initialization
      */
@@ -935,7 +948,7 @@ ImportObjCmd(clientData, interp, objc, objv)
     SSL_CTX_set_info_callback(statePtr->ctx, InfoCallback);
 
     /* Create Tcl_Channel BIO Handler */
-    statePtr->p_bio	= BIO_new_tcl(statePtr, BIO_CLOSE);
+    statePtr->p_bio	= BIO_new_tcl(statePtr, BIO_NOCLOSE);
     statePtr->bio	= BIO_new(BIO_f_ssl());
 
     if (server) {
@@ -950,9 +963,11 @@ ImportObjCmd(clientData, interp, objc, objv)
     /*
      * End of SSL Init
      */
+    dprintf("Returning %s", Tcl_GetChannelName(statePtr->self));
     Tcl_SetResult(interp, (char *) Tcl_GetChannelName(statePtr->self),
 	    TCL_VOLATILE);
     return TCL_OK;
+    	clientData = clientData;
 }
 
 /*
@@ -980,6 +995,8 @@ UnimportObjCmd(clientData, interp, objc, objv)
 {
     Tcl_Channel chan;		/* The channel to set a mode on. */
 
+    dprintf("Called");
+
     if (objc != 2) {
 	Tcl_WrongNumArgs(interp, 1, objv, "channel");
 	return TCL_ERROR;
@@ -990,12 +1007,10 @@ UnimportObjCmd(clientData, interp, objc, objv)
 	return TCL_ERROR;
     }
 
-    if (channelTypeVersion == TLS_CHANNEL_VERSION_2) {
-	/*
-	 * Make sure to operate on the topmost channel
-	 */
-	chan = Tcl_GetTopChannel(chan);
-    }
+    /*
+     * Make sure to operate on the topmost channel
+     */
+    chan = Tcl_GetTopChannel(chan);
 
     if (Tcl_GetChannelType(chan) != Tls_ChannelType()) {
 	Tcl_AppendResult(interp, "bad channel \"", Tcl_GetChannelName(chan),
@@ -1008,6 +1023,7 @@ UnimportObjCmd(clientData, interp, objc, objv)
     }
 
     return TCL_OK;
+    	clientData = clientData;
 }
 
 /*
@@ -1025,7 +1041,7 @@ UnimportObjCmd(clientData, interp, objc, objv)
  */
 
 static SSL_CTX *
-CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
+CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers, DHparams)
     State *statePtr;
     int proto;
     char *key;
@@ -1033,6 +1049,7 @@ CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
     char *CAdir;
     char *CAfile;
     char *ciphers;
+    char *DHparams;
 {
     Tcl_Interp *interp = statePtr->interp;
     SSL_CTX *ctx = NULL;
@@ -1041,6 +1058,8 @@ CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
     int off = 0;
     const SSL_METHOD *method;
 
+    dprintf("Called");
+
     if (!proto) {
 	Tcl_AppendResult(interp, "no valid protocol selected", NULL);
 	return (SSL_CTX *)0;
@@ -1123,7 +1142,7 @@ CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
 #endif
 	break;
     }
-
+    
     ctx = SSL_CTX_new (method);
     
     SSL_CTX_set_app_data( ctx, (VOID*)interp);	/* remember the interpreter */
@@ -1141,9 +1160,41 @@ CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
     SSL_CTX_set_default_passwd_cb_userdata(ctx, (void *)statePtr);
 #endif
 
-#ifndef NO_DH
+    /* read a Diffie-Hellman parameters file, or use the built-in one */
+#ifdef OPENSSL_NO_DH
+    if (DHparams != NULL) {
+	Tcl_AppendResult(interp,
+	    "DH parameter support not available", (char *) NULL);
+	SSL_CTX_free(ctx);
+	return (SSL_CTX *)0;
+    }
+#else
     {
-	DH* dh = get_dh512();
+	DH* dh;
+	if (DHparams != NULL) {
+	    BIO *bio;
+	    Tcl_DStringInit(&ds);
+	    bio = BIO_new_file(F2N(DHparams, &ds), "r");
+	    if (!bio) {
+		Tcl_DStringFree(&ds);
+		Tcl_AppendResult(interp,
+		    "Could not find DH parameters file", (char *) NULL);
+		SSL_CTX_free(ctx);
+		return (SSL_CTX *)0;
+	    }
+	    
+	    dh = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
+	    BIO_free(bio);
+	    Tcl_DStringFree(&ds);
+	    if (!dh) {
+		Tcl_AppendResult(interp,
+		    "Could not read DH parameters from file", (char *) NULL);
+		SSL_CTX_free(ctx);
+		return (SSL_CTX *)0;
+	    }
+	} else {
+	    dh = get_dhParams();
+	}
 	SSL_CTX_set_tmp_dh(ctx, dh);
 	DH_free(dh);
     }
@@ -1217,7 +1268,14 @@ CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
 	return (SSL_CTX *)0;
 #endif
     }
-    SSL_CTX_set_client_CA_list(ctx, SSL_load_client_CA_file( F2N(CAfile, &ds) ));
+
+    /* https://sourceforge.net/p/tls/bugs/57/ */
+    if ( CAfile != NULL ) {
+        STACK_OF(X509_NAME) *certNames = SSL_load_client_CA_file( F2N(CAfile, &ds) );
+	if ( certNames != NULL ) { 
+	    SSL_CTX_set_client_CA_list(ctx, certNames );
+	}
+    }
 
     Tcl_DStringFree(&ds);
     Tcl_DStringFree(&ds1);
@@ -1251,6 +1309,8 @@ StatusObjCmd(clientData, interp, objc, objv)
     char *channelName, *ciphers;
     int mode;
 
+    dprintf("Called");
+
     switch (objc) {
 	case 2:
 	    channelName = Tcl_GetStringFromObj(objv[1], NULL);
@@ -1271,12 +1331,10 @@ StatusObjCmd(clientData, interp, objc, objv)
     if (chan == (Tcl_Channel) NULL) {
 	return TCL_ERROR;
     }
-    if (channelTypeVersion == TLS_CHANNEL_VERSION_2) {
-	/*
-	 * Make sure to operate on the topmost channel
-	 */
-	chan = Tcl_GetTopChannel(chan);
-    }
+    /*
+     * Make sure to operate on the topmost channel
+     */
+    chan = Tcl_GetTopChannel(chan);
     if (Tcl_GetChannelType(chan) != Tls_ChannelType()) {
 	Tcl_AppendResult(interp, "bad channel \"", Tcl_GetChannelName(chan),
 		"\": not a TLS channel", NULL);
@@ -1309,6 +1367,7 @@ StatusObjCmd(clientData, interp, objc, objv)
     }
     Tcl_SetObjResult( interp, objPtr);
     return TCL_OK;
+    	clientData = clientData;
 }
 
 /*
@@ -1333,10 +1392,15 @@ VersionObjCmd(clientData, interp, objc, objv)
 {
     Tcl_Obj *objPtr;
 
+    dprintf("Called");
+
     objPtr = Tcl_NewStringObj(OPENSSL_VERSION_TEXT, -1);
 
     Tcl_SetObjResult(interp, objPtr);
     return TCL_OK;
+    	clientData = clientData;
+    	objc = objc;
+    	objv = objv;
 }
 
 /*
@@ -1363,6 +1427,8 @@ MiscObjCmd(clientData, interp, objc, objv)
     enum command { C_REQ, C_DUMMY };
     int cmd;
 
+    dprintf("Called");
+
     if (objc < 2) {
 	Tcl_WrongNumArgs(interp, 1, objv, "subcommand ?args?");
 	return TCL_ERROR;
@@ -1464,13 +1530,13 @@ MiscObjCmd(clientData, interp, objc, objv)
 		
 		name=X509_get_subject_name(cert);
 
-		X509_NAME_add_entry_by_txt(name,"C", MBSTRING_ASC, k_C, -1, -1, 0);
-		X509_NAME_add_entry_by_txt(name,"ST", MBSTRING_ASC, k_ST, -1, -1, 0);
-		X509_NAME_add_entry_by_txt(name,"L", MBSTRING_ASC, k_L, -1, -1, 0);
-		X509_NAME_add_entry_by_txt(name,"O", MBSTRING_ASC, k_O, -1, -1, 0);
-		X509_NAME_add_entry_by_txt(name,"OU", MBSTRING_ASC, k_OU, -1, -1, 0);
-		X509_NAME_add_entry_by_txt(name,"CN", MBSTRING_ASC, k_CN, -1, -1, 0);
-		X509_NAME_add_entry_by_txt(name,"Email", MBSTRING_ASC, k_Email, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"C", MBSTRING_ASC, (unsigned char *) k_C, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"ST", MBSTRING_ASC, (unsigned char *) k_ST, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"L", MBSTRING_ASC, (unsigned char *) k_L, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"O", MBSTRING_ASC, (unsigned char *) k_O, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"OU", MBSTRING_ASC, (unsigned char *) k_OU, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"CN", MBSTRING_ASC, (unsigned char *) k_CN, -1, -1, 0);
+		X509_NAME_add_entry_by_txt(name,"Email", MBSTRING_ASC, (unsigned char *) k_Email, -1, -1, 0);
 
 		X509_set_subject_name(cert,name);
 
@@ -1499,6 +1565,7 @@ MiscObjCmd(clientData, interp, objc, objv)
 	break;
     }
     return TCL_OK;
+    	clientData = clientData;
 }
 
 /*
@@ -1522,6 +1589,8 @@ Tls_Free( char *blockPtr )
 {
     State *statePtr = (State *)blockPtr;
 
+    dprintf("Called");
+
     Tls_Clean(statePtr);
     ckfree(blockPtr);
 }
@@ -1544,13 +1613,12 @@ Tls_Free( char *blockPtr )
  *
  *-------------------------------------------------------------------
  */
-void
-Tls_Clean(State *statePtr)
-{
+void Tls_Clean(State *statePtr) {
+    dprintf("Called");
+
     /*
      * we're assuming here that we're single-threaded
      */
-
     if (statePtr->timer != (Tcl_TimerToken) NULL) {
 	Tcl_DeleteTimerHandler(statePtr->timer);
 	statePtr->timer = NULL;
@@ -1558,12 +1626,12 @@ Tls_Clean(State *statePtr)
 
     if (statePtr->bio) {
 	/* This will call SSL_shutdown. Bug 1414045 */
-	dprintf(stderr, "BIO_free_all(%p)\n", statePtr->bio);
+	dprintf("BIO_free_all(%p)", statePtr->bio);
 	BIO_free_all(statePtr->bio);
 	statePtr->bio = NULL;
     }
     if (statePtr->ssl) {
-	dprintf(stderr, "SSL_free(%p)\n", statePtr->ssl);
+	dprintf("SSL_free(%p)", statePtr->ssl);
 	SSL_free(statePtr->ssl);
 	statePtr->ssl = NULL;
     }
@@ -1579,6 +1647,8 @@ Tls_Clean(State *statePtr)
 	Tcl_DecrRefCount(statePtr->password);
 	statePtr->password = NULL;
     }
+
+    dprintf("Returning");
 }
 
 /*
@@ -1597,72 +1667,47 @@ Tls_Clean(State *statePtr)
  *-------------------------------------------------------------------
  */
 
-int
-Tls_Init(Tcl_Interp *interp)		/* Interpreter in which the package is
-					 * to be made available. */
-{
-    int major, minor, patchlevel, release;
+int Tls_Init(Tcl_Interp *interp) {
+	const char tlsTclInitScript[] = {
+#include "tls.tcl.h"
+            0x00
+	};
 
-    /*
-     * The original 8.2.0 stacked channel implementation (and the patch
-     * that preceded it) had problems with scalability and robustness.
-     * These were address in 8.3.2 / 8.4a2, so we now require that as a
-     * minimum for TLS 1.4+.  We only support 8.2+ now (8.3.2+ preferred).
-     */
-    if (
+        dprintf("Called");
+
+	/*
+	 * We only support Tcl 8.4 or newer
+	 */
+	if (
 #ifdef USE_TCL_STUBS
-	Tcl_InitStubs(interp, "8.2", 0)
+	    Tcl_InitStubs(interp, "8.4", 0)
 #else
-	Tcl_PkgRequire(interp, "Tcl", "8.2", 0)
+	    Tcl_PkgRequire(interp, "Tcl", "8.4", 0)
 #endif
-	== NULL) {
-	return TCL_ERROR;
-    }
+	     == NULL) {
+		return TCL_ERROR;
+	}
 
-    /*
-     * Get the version so we can runtime switch on available functionality.
-     * TLS should really only be used in 8.3.2+, but the other works for
-     * some limited functionality, so an attempt at support is made.
-     */
-    Tcl_GetVersion(&major, &minor, &patchlevel, &release);
-    if ((major > 8) || ((major == 8) && ((minor > 3) || ((minor == 3) &&
-	    (release == TCL_FINAL_RELEASE) && (patchlevel >= 2))))) {
-	/* 8.3.2+ */
-	channelTypeVersion = TLS_CHANNEL_VERSION_2;
-    } else {
-	/* 8.2.0 - 8.3.1 */
-	channelTypeVersion = TLS_CHANNEL_VERSION_1;
-    }
+	if (TlsLibInit(0) != TCL_OK) {
+		Tcl_AppendResult(interp, "could not initialize SSL library", NULL);
+		return TCL_ERROR;
+	}
 
-    if (TlsLibInit() != TCL_OK) {
-	Tcl_AppendResult(interp, "could not initialize SSL library", NULL);
-	return TCL_ERROR;
-    }
+	Tcl_CreateObjCommand(interp, "tls::ciphers", CiphersObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	Tcl_CreateObjCommand(interp, "tls::handshake", HandshakeObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	Tcl_CreateObjCommand(interp, "tls::import", ImportObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	Tcl_CreateObjCommand(interp, "tls::unimport", UnimportObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	Tcl_CreateObjCommand(interp, "tls::status", StatusObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	Tcl_CreateObjCommand(interp, "tls::version", VersionObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	Tcl_CreateObjCommand(interp, "tls::misc", MiscObjCmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
 
-    Tcl_CreateObjCommand(interp, "tls::ciphers", CiphersObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
+	if (interp) {
+		Tcl_Eval(interp, tlsTclInitScript);
+	}
 
-    Tcl_CreateObjCommand(interp, "tls::handshake", HandshakeObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
-
-    Tcl_CreateObjCommand(interp, "tls::import", ImportObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
-
-    Tcl_CreateObjCommand(interp, "tls::unimport", UnimportObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
-
-    Tcl_CreateObjCommand(interp, "tls::status", StatusObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
-
-    Tcl_CreateObjCommand(interp, "tls::version", VersionObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
-
-    Tcl_CreateObjCommand(interp, "tls::misc", MiscObjCmd,
-	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
-
-    return Tcl_PkgProvide(interp, PACKAGE_NAME, PACKAGE_VERSION);
+	return(Tcl_PkgProvide(interp, "tls", PACKAGE_VERSION));
 }
-
+
 /*
  *------------------------------------------------------*
  *
@@ -1682,13 +1727,11 @@ Tls_Init(Tcl_Interp *interp)		/* Interpreter in which the package is
  *------------------------------------------------------*
  */
 
-int
-Tls_SafeInit (Tcl_Interp* interp)
-{
-    return Tls_Init (interp);
+int Tls_SafeInit(Tcl_Interp *interp) {
+	dprintf("Called");
+	return(Tls_Init(interp));
 }
 
-
 /*
  *------------------------------------------------------*
  *
@@ -1706,67 +1749,104 @@ Tls_SafeInit (Tcl_Interp* interp)
  *
  *------------------------------------------------------*
  */
-static int
-TlsLibInit ()
-{
-    int i;
-    char rnd_seed[16] = "GrzSlplKqUdnnzP!";	/* 16 bytes */
+static int TlsLibInit(int uninitialize) {
+	static int initialized = 0;
+	int status = TCL_OK;
 #if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
-    size_t num_locks;
-#endif
-    int status=TCL_OK;
-
-    if (!initialized) {
-	Tcl_MutexLock(&init_mx);
-	if (!initialized) {
-	    initialized = 1;
-
-	    if (CRYPTO_set_mem_functions((void *(*)(size_t))Tcl_Alloc,
-					 (void *(*)(void *, size_t))Tcl_Realloc,
-					 (void(*)(void *))Tcl_Free) == 0) {
-	       /* Not using Tcl's mem functions ... not critical */
-	    }
-
-#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
-	    /* should we consider allocating mutexes? */
-	    num_locks = CRYPTO_num_locks();
-	    if (num_locks > CRYPTO_NUM_LOCKS) {
-		status=TCL_ERROR;
-		goto done;
-	    }
-
-	    CRYPTO_set_locking_callback(CryptoThreadLockCallback);
-	    CRYPTO_set_id_callback(CryptoThreadIdCallback);
+	size_t num_locks;
 #endif
 
-	    if (SSL_library_init() != 1) {
-	    	status=TCL_ERROR;
-		goto done;
-	    }
-	    SSL_load_error_strings();
-	    ERR_load_crypto_strings();
+	if (uninitialize) {
+		if (!initialized) {
+			dprintf("Asked to uninitialize, but we are not initialized");
 
-	    /*
-	     * Seed the random number generator in the SSL library,
-	     * using the do/while construct because of the bug note in the
-	     * OpenSSL FAQ at http://www.openssl.org/support/faq.html#USER1
-	     *
-	     * The crux of the problem is that Solaris 7 does not have a 
-	     * /dev/random or /dev/urandom device so it cannot gather enough
-	     * entropy from the RAND_seed() when TLS initializes and refuses
-	     * to go further. Earlier versions of OpenSSL carried on regardless.
-	     */
-	    srand((unsigned int) time((time_t *) NULL));
-	    do {
-		for (i = 0; i < 16; i++) {
-		    rnd_seed[i] = 1 + (char) (255.0 * rand()/(RAND_MAX+1.0));
+			return(TCL_OK);
 		}
-		RAND_seed(rnd_seed, sizeof(rnd_seed));
-	    } while (RAND_status() != 1);
-	}
-    	done:
 
+		dprintf("Asked to uninitialize");
+
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
+		Tcl_MutexLock(&init_mx);
+
+		CRYPTO_set_locking_callback(NULL);
+		CRYPTO_set_id_callback(NULL);
+
+		if (locks) {
+			free(locks);
+			locks = NULL;
+			locksCount = 0;
+		}
+#endif
+		initialized = 0;
+
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
+		Tcl_MutexUnlock(&init_mx);
+#endif
+
+		return(TCL_OK);
+	}
+
+	if (initialized) {
+		dprintf("Called, but using cached value");
+		return(status);
+	}
+
+	dprintf("Called");
+
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
+	Tcl_MutexLock(&init_mx);
+#endif
+	initialized = 1;
+
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
+	num_locks = CRYPTO_num_locks();
+	locksCount = num_locks;
+	locks = malloc(sizeof(*locks) * num_locks);
+	memset(locks, 0, sizeof(*locks) * num_locks);
+
+	CRYPTO_set_locking_callback(CryptoThreadLockCallback);
+	CRYPTO_set_id_callback(CryptoThreadIdCallback);
+#endif
+
+	if (SSL_library_init() != 1) {
+		status = TCL_ERROR;
+		goto done;
+	}
+
+	SSL_load_error_strings();
+	ERR_load_crypto_strings();
+
+	BIO_new_tcl(NULL, 0);
+
+#if 0
+	/*
+	 * XXX:TODO: Remove this code and replace it with a check
+	 * for enough entropy and do not try to create our own
+	 * terrible entropy
+	 */
+    /*
+     * Seed the random number generator in the SSL library,
+     * using the do/while construct because of the bug note in the
+     * OpenSSL FAQ at http://www.openssl.org/support/faq.html#USER1
+     *
+     * The crux of the problem is that Solaris 7 does not have a 
+     * /dev/random or /dev/urandom device so it cannot gather enough
+     * entropy from the RAND_seed() when TLS initializes and refuses
+     * to go further. Earlier versions of OpenSSL carried on regardless.
+     */
+    srand((unsigned int) time((time_t *) NULL));
+    do {
+	for (i = 0; i < 16; i++) {
+	    rnd_seed[i] = 1 + (char) (255.0 * rand()/(RAND_MAX+1.0));
+	}
+	RAND_seed(rnd_seed, sizeof(rnd_seed));
+    } while (RAND_status() != 1);
+#endif
+
+done:
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
 	Tcl_MutexUnlock(&init_mx);
-    }
-    return status;
+#endif
+
+	return(status);
 }
